﻿using FactPDA.Models;
using FactPOP.Models;
using FactPOP.Win.Utils;
using Newtonsoft.Json;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Data;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Http;
using System.Net.Sockets;
using System.Reflection;
using System.Runtime.Serialization.Formatters.Binary;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using System.Xml;
using System.Xml.Serialization;

namespace FactPDA
{
    public static class DataCommon
    {
        public static ReturnData ProcCall(List<SendDataV2> para)
        {
            using (ReturnData returnData = new ReturnData())
            {
                DataRequest req = null;
                DataResult res = null;
                try
                {
                    req = new DataRequest();
                    req.ProcParam.sendValueV2 = para;
                    res = ServerProcRequest(req);
                    if (!res.IsError)
                    {
                        if (res.Data == null || res.Data.GetData(0) == null && res.Data.GetData(0).Rows.Count < 1)
                        {
                            returnData.success = false;
                            returnData.errorMessage = "데이터가 없습니다.";
                        }
                        else
                        {
                            returnData.errorMessages = res.Data.GetDataErrorMessage(null, true);
                            returnData.data = res.Data;
                            if (returnData.errorMessages.Count == 0)
                            {
                                returnData.success = true;
                            }
                            else
                            {
                                returnData.success = false;
                                returnData.errorMessage = string.Join(Environment.NewLine, returnData.errorMessages);
                            }
                        }
                    }
                    else
                    {
                        returnData.success = false;
                        returnData.errorMessage = res.Message;
                    }
                }
                catch (Exception ex)
                {
                    returnData.success = false;
                    returnData.errorMessage = ex.Message;
                }
                finally
                {
                    if (req != null) { req.Dispose(); req = null; }
                    if (res != null) { res.Dispose(); res = null; }
                }
                return returnData;
            }

        }

        public static List<string> GetDataErrorMessage(this DataStore<DataTable> data, int? tableIndex = null, bool allRowMessage = false)
        {
            List<string> errorMessages = new List<string>();
            if (data != null)
            {
                if (tableIndex == null)
                {
                    foreach (KeyValuePair<string, DataTable> kv in data)
                    {
                        if (kv.Value != null && kv.Value.Rows.Count > 0 && kv.Value.Columns.Contains("ERR_NO") && kv.Value.Columns.Contains("ERR_MSG"))
                        {
                            if (allRowMessage == true)
                            {
                                foreach (DataRow row in kv.Value.Rows)
                                {
                                    if (Convert.ToInt32(row["ERR_NO"]) != 0)
                                    {
                                        errorMessages.Add(row["ERR_MSG"].ToString());
                                    }
                                }
                            }
                            else
                            {
                                if (kv.Value.Columns.Contains("ERR_NO"))
                                {
                                    errorMessages.Add(kv.Value.Rows[0]["ERR_MSG"].ToString());
                                }
                                else
                                {
                                    errorMessages.Add(kv.Value.Rows[0]["ERR_NO"].ToString());
                                }
                            }
                        }
                        else if (kv.Value != null && kv.Value.Rows.Count > 0 && (kv.Value.Columns.Contains("ERROR_NUMBER") || kv.Value.Columns.Contains("ERROR_CODE")) && kv.Value.Columns.Contains("ERROR_MESSAGE"))
                        {
                            if (allRowMessage == true)
                            {
                                foreach (DataRow row in kv.Value.Rows)
                                {
                                    if (kv.Value.Columns.Contains("ERROR_NUMBER"))
                                    {
                                        if (Convert.ToInt32(row["ERROR_NUMBER"]) != 0)
                                        {
                                            errorMessages.Add(row["ERROR_MESSAGE"].ToString());
                                        }
                                    }
                                    else
                                    {
                                        if (Convert.ToInt32(row["ERROR_CODE"]) != 0)
                                        {
                                            errorMessages.Add(row["ERROR_MESSAGE"].ToString());
                                        }
                                    }

                                }
                            }
                            else
                            {
                                if (kv.Value.Columns.Contains("ERROR_NUMBER"))
                                {
                                    errorMessages.Add(kv.Value.Rows[0]["ERROR_MESSAGE"].ToString());
                                }
                                else
                                {
                                    errorMessages.Add(kv.Value.Rows[0]["ERROR_CODE"].ToString());
                                }
                            }

                        }
                    }
                }
                else
                {
                    int tIndex = (int)tableIndex;

                    DataTable dt = data[data.ElementAt(tIndex).Key];
                    if (dt.Columns.Contains("ERR_NO") && dt.Columns.Contains("ERR_MSG"))
                    {
                        if (allRowMessage)
                        {
                            foreach (DataRow row in dt.Rows)
                            {
                                if (dt.Columns.Contains("ERR_NO"))
                                {
                                    if (Convert.ToInt32(row["ERR_NO"]) != 0)
                                    {
                                        errorMessages.Add(row["ERR_MSG"].ToString());
                                    }
                                }
                            }
                        }
                        else
                        {
                            if (dt.Columns.Contains("ERR_NO"))
                            {
                                if (Convert.ToInt32(dt.Rows[0]["ERR_NO"]) != 0)
                                {
                                    errorMessages.Add(dt.Rows[0]["ERR_MSG"].ToString());
                                }
                            }
                        }
                    }
                    else if ((dt.Columns.Contains("ERROR_NUMBER") || dt.Columns.Contains("ERROR_CODE")) && dt.Columns.Contains("ERROR_MESSAGE"))
                    {
                        if (allRowMessage)
                        {
                            foreach (DataRow row in dt.Rows)
                            {
                                if (dt.Columns.Contains("ERROR_NUMBER"))
                                {
                                    if (Convert.ToInt32(row["ERROR_NUMBER"]) != 0)
                                    {
                                        errorMessages.Add(row["ERROR_MESSAGE"].ToString());
                                    }
                                }
                                else
                                {
                                    if (Convert.ToInt32(row["ERROR_CODE"]) != 0)
                                    {
                                        errorMessages.Add(row["ERROR_MESSAGE"].ToString());
                                    }
                                }

                            }
                        }
                        else
                        {
                            if (dt.Columns.Contains("ERROR_NUMBER"))
                            {
                                if (Convert.ToInt32(dt.Rows[0]["ERROR_NUMBER"]) != 0)
                                {
                                    errorMessages.Add(dt.Rows[0]["ERROR_MESSAGE"].ToString());
                                }
                            }
                            else
                            {
                                if (Convert.ToInt32(dt.Rows[0]["ERROR_CODE"]) != 0)
                                {
                                    errorMessages.Add(dt.Rows[0]["ERROR_MESSAGE"].ToString());
                                }
                            }

                        }

                    }
                }
            }

#if DEBUG
            if (data != null)
            {
                foreach (var kv in data)
                {
                    DataTable dt = kv.Value;

                    if ((dt.Columns.Contains("ERROR_NUMBER") || dt.Columns.Contains("ERROR_CODE")) && dt.Columns.Contains("ERROR_POSITION"))
                    {
                        foreach (DataRow row in dt.Rows)
                        {
                            if (dt.Columns.Contains("ERROR_NUMBER"))
                            {
                                if (Convert.ToInt32(row["ERROR_NUMBER"]) != 0)
                                {
                                    errorMessages.Add(row["ERROR_POSITION"].ToString());
                                }
                            }
                            else
                            {
                                if (Convert.ToInt32(row["ERROR_CODE"]) != 0)
                                {
                                    errorMessages.Add(row["ERROR_POSITION"].ToString());
                                }
                            }
                        }
                    }
                }
            }
#endif
            return errorMessages.Distinct().ToList();
        }

        public static DataResult ServerProcRequest(DataRequest request)
        {
            try
            {
                return (new ProcHttpClient()).Execute(request);
            }
            catch (Exception ex)
            {
                throw (ex);
            }
        }
    }

    public class ReturnData : IDisposable
    {
        public List<string> errorMessages;
        public string errorMessage;
        public DataStore<DataTable> data;
        public bool success;

        void IDisposable.Dispose() { }
        public void Dispose() { }
    }

    public class ProcHttpClient : HttpClientInit
    {
        //웹 접근 URI
        const string webapiUri = @"api/Proc";

        //---------------------------------------------------------------------
        // 일반적인 실행 메소드
        //---------------------------------------------------------------------
        #region Execute
        public DataResult Execute(DataRequest request)
        {
            try
            {
                request.CallCls = "PDA";

                //-------------------------------------------------------------
                // 글로벌 변수로 선언된 값을 변경한다.
                //-------------------------------------------------------------
                if (request.IsReal.Equals("AUTO"))
                    request.IsReal = GlobalVar.IsReal;

                //-------------------------------------------------------------
                // DataRequest 객체를 Json형식으로 변환하여 
                // Post방식으로 서버에 요청한다.
                //-------------------------------------------------------------
                //JsonMediaTypeFormatter jsonFormatter = new JsonMediaTypeFormatter();


                #region OtherValue
                // TODO : 공통 파라미터 추가

                Dictionary<string, object> OtherValue = new Dictionary<string, object>();
                OtherValue.Add("REG_DWGP_CD", GlobalVar.CorpCd);
                OtherValue.Add("REG_EMP_NO", GlobalVar.EmpNo);

                OtherValue.Add("R_CORP_CD", GlobalVar.CorpCd);
                OtherValue.Add("R_EMP_NO", GlobalVar.EmpNo);

                request.OtherValue = OtherValue;
                #endregion

                //HttpContent content = new ObjectContent<DataRequest>(request, jsonFormatter);
                request.JsonData = JsonConvert.SerializeObject(request);
                var jsonRequest = JsonConvert.SerializeObject(request);
                var httpContent = new StringContent(jsonRequest, System.Text.Encoding.UTF8, "application/json");


                //var resp = Init().PostAsync(webapiUri, content).Result;

                var resp = Init().PostAsync(webapiUri, httpContent).Result;

                //var resp = Init().PostA(webapiUri, )
                //var resp = Init().PostAsJsonAsync<DataRequest>(webapiUri, request).Result;
                if (resp.IsSuccessStatusCode)
                {
                    //var result = resp.Content.ReadAsAsync<DataResult>().Result;
                    //var result = resp.Content.ReadAsAsync<DataResult>().Result;
                    //string ttt = resp.Content.ReadAsStringAsync().Result;


                    var resultContent = resp.Content.ReadAsStringAsync().Result;

                    var result = JsonConvert.DeserializeObject<DataResult>(resultContent);

                    //var result = JsonConvert.DeserializeObject(resp.Content.ReadAsStringAsync().Result) as DataResult;

                    //var result = new DataResult();
                    ////오류내역: Unexpected end when deserializing array --> 반환되는 데이터테이블이 널인 경우 오류 발생
                    //var output = JsonConvert.DeserializeObject<List<DataTable>>(result.Data);

                    //서버에서 오류가 발생하여 Catch되어 반환된 경우에는 result.Data가 NULL이기 때문에 
                    //Null이 아닌 경우에만 아래의 내용을 처리하도록 if문 추가함 By Woo Jong Ho 2014.03.07
                    if (result.Data != null)
                    {
                        for (int i = 0; i < result.Data.Count; i++)
                        {
                            if (result.GetData(i).Columns.Count == 0 && result.GetData(i).Rows.Count == 0)
                            {
                                foreach (var item in result.Schema[i])
                                {
                                    result.GetData(i).Columns.Add(item.Key, item.Value);
                                }
                            }

                            #region DataType 보정
                            Dictionary<string, Type> schema = null;
                            if (result.Schema.Count > 0 && result.Schema[i] is Dictionary<string, Type>)
                            {
                                schema = result.Schema[i] as Dictionary<string, Type>;
                            }

                            if (schema != null)
                            {
                                var columns = result.GetData(i).Columns.Cast<DataColumn>().Where(col => schema.ContainsKey(col.ColumnName) && col.DataType != schema[col.ColumnName]).ToList();

                                List<KeyValuePair<string, Type>> dataColumnConvertColumn = new List<KeyValuePair<string, Type>>();

                                //foreach (DataColumn column in columns)
                                //{
                                //    dataColumnConvertColumn.Add(new KeyValuePair<string, Type>(column.ColumnName, schema[column.ColumnName]));
                                //}
                                result.SetData(i, result.GetData(i).GetDataTableColumnConvert(dataColumnConvertColumn));
                            }
                            #endregion

                        }

                        string[] dateColumn = { "_DTM", "_DT", "_DATE" };
                        for (int i = 0; i < result.Data.Count; i++)
                        {
                            // Column 잡아 돌림
                            // 조건1 타입이 String
                            // 조건2 컬럼명의 끝문자열이 dateColumn List에 포함되는지
                            var columnNames = result.GetData(i).Columns.OfType<DataColumn>()
                                .Where(col => col.DataType == typeof(String))
                                .Select(col => new { columnName = col.ColumnName, endName = "_" + (col.ColumnName.Split('_').Length > 1 ? col.ColumnName.Split('_')[col.ColumnName.Split('_').Length - 1] : col.ColumnName) })
                                .Where(column => dateColumn.Contains(column.endName));


                            //foreach (var column in columnNames)
                            //{
                            //    // Row 잡아 돌림
                            //    // 조건3 모든 데이터가 공백 또는 DBNull 인지
                            //    if (result.GetData(i).Rows.Cast<DataRow>().AsEnumerable().Where(row => row[column.columnName] == DBNull.Value || string.IsNullOrEmpty(row[column.columnName].ToString())).Count() == result.GetData(i).Rows.Count)
                            //    {
                            //        // 그렇다라면 타입 바꿈
                            //        List<KeyValuePair<string, Type>> dataColumnConvertColumn = new List<KeyValuePair<string, Type>>();
                            //        dataColumnConvertColumn.Add(new KeyValuePair<string, Type>(column.columnName, typeof(DateTime)));
                            //        result.SetData(i, result.GetData(i).GetDataTableColumnConvert(dataColumnConvertColumn));
                            //    }
                            //}

                        }
                    }
                    return result;
                }
                else
                {
                    throw new Exception(resp.StatusCode.ToString() + Environment.NewLine +
                                        resp.RequestMessage.ToString() + Environment.NewLine +
                                        resp.ReasonPhrase.ToArray());
                }
            }
            catch (TaskCanceledException)
            {
                throw new Exception("RequestTimeOut");
            }
            catch (WebException we)
            {
                throw new Exception(we.Message);
            }
            catch (Exception ex)
            {
                throw new Exception((new ErrorHandler()).GetMessage(ex));
            }
        }
        #endregion
    }

    public class HttpClientInit
    {
        //----------------------------------------------------------------------------
        // HttpClient 생성 및 객체 반환
        //----------------------------------------------------------------------------     
        public HttpClient Init()
        {
            try
            {
                HttpClient client = new HttpClient();
                if (GlobalVar.WebServerMode.Equals("REAL"))
                    client.BaseAddress = new Uri(GlobalVar.RealServerURI);
                else if (GlobalVar.WebServerMode.Equals("DEBUG"))
                    client.BaseAddress = new Uri(GlobalVar.DebugServerURI);
                else
                    client.BaseAddress = new Uri(GlobalVar.TestServerURI);

                client.DefaultRequestHeaders.Accept.Add(new System.Net.Http.Headers.MediaTypeWithQualityHeaderValue("application/json"));
                return client;
            }
            catch (Exception)
            {
                throw;
            }
        }
    }

    public static class FuncLib
    {
        public static string GetIp()
        {
            IPHostEntry ipHost = Dns.GetHostEntry(Dns.GetHostName());
            string ip = String.Empty;
            for (int i = 0; i < ipHost.AddressList.Length; i++)
            {
                if (ipHost.AddressList[i].AddressFamily == AddressFamily.InterNetwork)
                    ip = ipHost.AddressList[i].ToString();
            }
            return ip;
        }

        public static bool YnToBool(this object input)
        {
            return input.ToString() == "Y" ? true : false;
        }

        public static bool YnToBool(this string input)
        {
            return input == "Y" ? true : false;
        }

        public static DataTable DataTableColumnInputData(this DataTable dt, string columnName, object data)
        {
            if (!dt.Columns.Contains(columnName))
            {
                dt.Columns.Add(columnName);
            }
            foreach (DataRow row in dt.Rows)
            {
                row[columnName] = data;
            }

            return dt;
        }

        public static ObservableCollection<T> ToObservableCollection<T>(this DataTable dt)
        {
            ObservableCollection<T> ts = new ObservableCollection<T>();

            PropertyInfo[] propertyInfos = typeof(T).GetProperties();

            foreach (DataRow row in dt.Rows)
            {
                T model = (T)Activator.CreateInstance(typeof(T));

                foreach (PropertyInfo info in propertyInfos)
                {
                    if (dt.Columns.Contains(info.Name))
                    {
                        ParaInfoSetValue(row, model, info);
                    }
                }

                ts.Add(model);
            }

            return ts;
        }

        public static void SetDataTableModel<T>(this ObservableCollection<T> ts, DataTable dt, bool isAppend = false)
        {
            if (!isAppend)
            {
                ts.Clear();
            }

            PropertyInfo[] propertyInfos = typeof(T).GetProperties();

            foreach (DataRow row in dt.Rows)
            {
                T model = (T)Activator.CreateInstance(typeof(T));

                foreach (PropertyInfo info in propertyInfos)
                {
                    if (dt.Columns.Contains(info.Name))
                    {
                        ParaInfoSetValue(row, model, info);
                    }
                }

                ts.Add(model);
            }

        }

        public static DataTable GetDataTable<T>(this ObservableCollection<T> ts)
        {

            DataTable dt = new DataTable();

            PropertyInfo[] propertyInfos = typeof(T).GetProperties();

            foreach (PropertyInfo info in propertyInfos)
            {
                dt.Columns.Add(info.Name, Type.GetType(info.PropertyType.FullName));
            }

            foreach (T item in ts)
            {
                DataRow newRow = dt.NewRow();

                foreach (PropertyInfo info in propertyInfos)
                {
                    //dt.Columns.Add(info.Name, Type.GetType(info.PropertyType.FullName));

                    if (dt.Columns.Contains(info.Name))
                    {
                        newRow[info.Name] = item.GetType().GetProperty(info.Name).GetValue(item);
                    }
                }


                dt.Rows.Add(newRow);
            }
            return dt;

        }

        public static ObservableCollection<T> Copy<T>(this ObservableCollection<T> ts)
        {
            ObservableCollection<T> tempItem = new ObservableCollection<T>();
            foreach (var item in ts)
            {
                tempItem.Add(item);
            }
            return tempItem;
        }

        public static string[] GetItemSourceText(this ObservableCollection<CODEHELP_CODE> data)
        {
            return data.AsEnumerable().Select(item => item.NM).ToArray();
        }

        public static object GetItemSourceValue(this ObservableCollection<CODEHELP_CODE> data, int index)
        {
            return data[index].ID;
        }
        private static void ParaInfoSetValue<T>(DataRow row, T model, PropertyInfo info)
        {
            //object value = GetDefaultValue(Type.GetType(info.PropertyType.FullName));

            Type type = Type.GetType(info.PropertyType.FullName);

            if (row[info.Name] == DBNull.Value)
            {
                //info.SetValue(model, value);
            }
            else
            {
                object value = ConvertValue(row[info.Name], type);
                info.SetValue(model, value);
                //info.SetValue(model, row[info.Name]);
            }
        }

        public static object ConvertValue(object value, Type type)
        {
            if (type == typeof(String))
            {
                return Convert.ToString(value);
            }
            else if (type == typeof(Decimal))
            {
                return Convert.ToDecimal(value);
            }
            else if (type == typeof(DateTime))
            {
                return Convert.ToDateTime(value);
            }
            else
            {
                //return value;
                if (!type.ToString().Contains("StatusObject"))
                {
                    return Convert.ChangeType(value, type);
                }
                else
                {
                    return null;
                }
                
            }
        }

        public static T ConvertValue<T>(object value)
        {
            return (T)Convert.ChangeType(value, typeof(T));
        }

        private static object GetDefaultValue(Type t)
        {
            if (t.IsValueType)
                return Activator.CreateInstance(t);

            return null;
        }
        public static DataTable GetColumnFilter(this DataTable dt, params string[] viewColumns)
        {
            foreach (var col in dt.Columns.Cast<DataColumn>().ToList())
            {
                if (!viewColumns.Contains(col.ColumnName))
                {
                    dt.Columns.Remove(col.ColumnName);
                }
            }

            return dt;
        }

        public static DataTable ConvertTypeDateToString(this DataTable dt)
        {
            foreach (DataColumn col in dt.Columns)
            {
                if (col.ColumnName != "DATA_REGDT")
                {
                    if (col.DataType == typeof(DateTime))
                    {
                        dt = dt.GetDataTableColumnConvert(col.ColumnName, typeof(string));
                    }
                }
            }
            return dt;
        }

        public static DataTable RowAdd(this DataTable dt)
        {
            dt.Rows.Add(dt.NewRow());
            return dt;
        }

        public static DataParam DataTableToDataParam(this DataTable dt)
        {
            DataParam param = new DataParam();

            if (dt.Rows.Count > 0)
            {
                foreach (DataColumn column in dt.Columns)
                {
                    param.Add(column.ColumnName, dt.Rows[0][column]);
                }
            }
            return param;
        }

        public static DataTable DataTableColumnNameChange(this DataTable dt, string fromColumnName, string toColumnName)
        {
            if (!dt.Columns.Contains(toColumnName))
            {
                dt.Columns.Add(toColumnName, dt.Columns[fromColumnName].DataType);
            }

            foreach (DataRow row in dt.Rows)
            {
                row[toColumnName] = row[fromColumnName];
            }
            return dt;
        }

        public static DataTable DataTableColumnInputData(this DataTable dt, Dictionary<string, object> inputData)
        {
            foreach (var item in inputData)
            {
                DataTableColumnInputData(dt, item.Key, item.Value);
            }
            return dt;
        }

        public static DataTable GetStateToDataTable(this DataTable dt, DataRowState state)
        {

            if (state == DataRowState.Deleted)
            {
                DataTable dtDelete = dt.Clone();

                foreach (DataRow row in dt.Rows.Cast<DataRow>().AsEnumerable().Where(row => row.RowState == state).ToList())
                {
                    DataRow newRow = dtDelete.NewRow();

                    foreach (DataColumn column in dt.Columns)
                    {
                        newRow[column.ColumnName] = row[column.ColumnName, DataRowVersion.Original];
                    }

                    dtDelete.Rows.Add(newRow);
                }
                return dtDelete;
            }
            else
            {
                var list = dt.Rows.Cast<DataRow>().AsEnumerable().Where(row => row.RowState == state).ToList();
                if (list.Count > 0)
                {
                    DataTable returnDt = list.First().Table.Clone();
                    foreach (DataRow row in list)
                    {
                        returnDt.Rows.Add(row.ItemArray);
                    }
                    return returnDt;
                }
                else
                {
                    return dt;
                }
            }
        }

        public static DataTable GetStateAddToDataTable(this DataTable dt)
        {
            DataTable dtNew = dt.Clone();
            dtNew.Columns.Add("ROWSTATE");

            if (!dtNew.Columns.Contains("R_CRUD"))
            {
                dtNew.Columns.Add("R_CRUD");
            }

            foreach (DataRow row in dt.Rows.Cast<DataRow>().AsEnumerable().Where(row => row.RowState == DataRowState.Added
                || row.RowState == DataRowState.Modified
                || row.RowState == DataRowState.Deleted).ToList())
            {
                if (row.RowState == DataRowState.Deleted)
                {
                    DataRow newRow = dtNew.NewRow();
                    foreach (DataColumn column in dt.Columns)
                    {
                        newRow[column.ColumnName] = row[column.ColumnName, DataRowVersion.Original];
                    }
                    newRow["ROWSTATE"] = "DELETE";
                    newRow["R_CRUD"] = "D10";
                    dtNew.Rows.Add(newRow);
                }
                else
                {
                    DataRow newRow = dtNew.NewRow();
                    newRow.ItemArray = row.ItemArray;
                    newRow["ROWSTATE"] = (row.RowState == DataRowState.Modified) ? "UPDATE" : "INSERT";
                    newRow["R_CRUD"] = (row.RowState == DataRowState.Modified) ? "U10" : "I10";
                    dtNew.Rows.Add(newRow);

                }
            }

            dtNew = dtNew.ConvertTypeDateToString();
            return dtNew;
        }

        public static DataTable GetDataTableByFields(this DataTable dt, List<string> fields)
        {
            //선택한 컬럼이 아닌 경우 삭제
            List<DataColumn> cols = dt.Columns.OfType<DataColumn>().Where(x => !fields.Contains(x.ColumnName)).ToList();
            foreach (DataColumn col in cols)
            {
                dt.Columns.Remove(col);
            }
            return dt;
        }

        /// <summary>
        /// DataTable의 Column DataType 을 변경합니다.
        /// </summary>
        /// <param name="dt">변경할 DataTable</param>
        /// <param name="columns">변경할 Column List</param>
        /// <returns>DataTable</returns>
        public static DataTable GetDataTableColumnConvert(this DataTable dt, List<KeyValuePair<string, Type>> columns)
        {
            try
            {
                DataTable dtNew = dt.Clone();

                foreach (var item in columns)
                {
                    dtNew.Columns[item.Key].DataType = item.Value;
                }

                foreach (DataRow row in dt.Rows)
                {
                    dtNew.ImportRow(row);
                }

                return dtNew;
            }
            catch
            {
                return dt;
            }
        }

        public static DataTable GetDataTableColumnConvert(this DataTable dt, string columnName, Type type)
        {
            try
            {
                DataTable dtNew = dt.Clone();
                dtNew.Columns[columnName].DataType = type;
                foreach (DataRow row in dt.Rows)
                {
                    dtNew.ImportRow(row);
                }
                return dtNew;
            }
            catch
            {
                return dt;
            }
        }

        public static DataTable GetDataTableDateToStringConvert(this DataTable dt, string dateFormat = "yyyy-MM-dd")
        {
            try
            {
                DataTable dtNew = new DataTable();
                List<KeyValuePair<string, Type>> listConvert = new List<KeyValuePair<string, Type>>();
                foreach (DataColumn col in dt.Columns)
                {
                    if (col.DataType == typeof(DateTime))
                    {
                        listConvert.Add(new KeyValuePair<string, Type>(col.ColumnName, typeof(String)));
                    }
                }

                dtNew = dt.GetDataTableColumnConvert(listConvert);
                foreach (var item in listConvert)
                {
                    foreach (DataRow row in dtNew.Rows)
                    {
                        if (row[item.Key] != DBNull.Value && row[item.Key] != null)
                        {
                            row[item.Key] = Convert.ToDateTime(row[item.Key]).ToString(dateFormat);
                        }

                    }
                }
                return dtNew;
            }
            catch
            {
                return dt;
            }
        }

        public static string GetDateTimeConvert(this string input, string format = "yyyy-MM-dd")
        {
            return input.Length == 8 ? DateTime.ParseExact(input, "yyyy-MM-dd", null).ToString(format) : DateTime.Parse(input).ToString(format);
        }

        public static string GetDateTimeConvert(this object input, string format = "yyyy-MM-dd")
        {
            if (input != null && input.GetType() == typeof(DateTime))
            {
                return ((DateTime)input).ToString(format);
            }
            else if (input == null || !(input is string))
            {
                return "";
            }
            else
            {
                return input.ToString().Length == 8 ? DateTime.ParseExact(input.ToString(), "yyyy-MM-dd", null).ToString(format) : DateTime.Parse(input.ToString()).ToString(format);
            }
        }

        public static DateTime GetStringToDateTime(this object input, string format = "yyyy-MM-dd")
        {
            if (input == null || !(input is string))
            {
                return DateTime.Now;
            }
            else
            {
                return DateTime.ParseExact(input.ToString(), format, null);
            }
        }

        public static string ToStringToLowerUpper(this string input, bool? upperYN)
        {
            if (upperYN != null)
            {
                if ((bool)upperYN)
                {
                    input = input.ToUpper();
                }
                else if ((bool)upperYN == false)
                {
                    input = input.ToLower();
                }
            }
            return input;
        }

        /// <summary>
        /// XML 포맷으로 직렬화 합니다.
        /// </summary>
        /// <param name="obj">직렬화할 Object</param>
        /// <returns>직렬화된 XML 문자열</returns>
        public static string Serialize(this Object obj)
        {
            XmlSerializer ser = new XmlSerializer(obj.GetType());
            System.Text.StringBuilder sb = new System.Text.StringBuilder();
            System.IO.StringWriter writer = new System.IO.StringWriter(sb);
            if (obj.GetType() == typeof(DataTable))
            {
                DataTable dtTemp = (DataTable)obj;
                if (String.IsNullOrEmpty(dtTemp.TableName))
                {
                    dtTemp.TableName = "NONAME";
                }
            }

            ser.Serialize(writer, obj);
            XmlDocument doc = new XmlDocument();
            doc.LoadXml(sb.ToString());
            string xml = doc.InnerXml;
            return xml;
        }

        /// <summary>
        /// 직렬화된 XML 문자열을 Object 역직렬화 합니다.
        /// </summary>
        /// <param name="xml">직렬화된 XML 문자열</param>
        /// <param name="objType">직렬화한 Object 의 타입</param>
        /// <example>DataTable dataTable = (DataTable)stringValue.DeSerialize(typeof(DataTable));</example>
        /// <returns>역직렬화된 Object</returns>
        public static object DeSerialize(this string xml, Type objType)
        {
            XmlDocument doc = new XmlDocument();
            doc.LoadXml(xml);
            XmlNodeReader reader = new XmlNodeReader(doc.DocumentElement);
            XmlSerializer ser = new XmlSerializer(objType);
            object obj = ser.Deserialize(reader);
            return obj;
        }

        public static DataTable GetDataTableValueChange(this DataTable dt, bool isDateToString8 = false)
        {
            foreach (DataRow row in dt.Rows)
            {
                if (isDateToString8)
                {
                    row.ItemArray.Where(cellValue => cellValue.GetType() == typeof(DateTime)).ToList().ForEach(delegate (object obj) { obj = DateTime.Parse(obj.ToString()).ToShortDateString(); });
                }
            }
            return dt;
        }

        public static Decimal GetNullToZero(this object obj)
        {
            if (obj == null || obj == DBNull.Value || obj.ToString() == "")
            {
                return 0;
            }
            else
            {
                return Convert.ToDecimal(obj);
            }
        }

        public static int GetIntNullToZero(this object obj)
        {
            if (obj == null || obj == DBNull.Value || obj.ToString() == "")
            {
                return 0;
            }
            else
            {
                return Convert.ToInt32(obj);
            }
        }

        public static int? GetIntNullToZeroCheck(this object obj)
        {
            if (obj == null || obj == DBNull.Value || obj.ToString() == "")
            {
                return null;
            }
            else
            {
                int n;
                bool isNumeric = int.TryParse(obj.ToString(), out n);
                if (isNumeric)
                {
                    return Convert.ToInt32(obj);
                }
                else
                {
                    return null;
                }
            }
        }

        public static int GetDeciamlToInt(this decimal input)
        {
            return Convert.ToInt32(input);
        }

        public static int GetDoubleToInt(this double input)
        {
            return Convert.ToInt32(input);
        }

        public static double GetDoubleNullToZero(this object obj)
        {
            if (obj == null || obj == DBNull.Value || obj.ToString() == "")
            {
                return 0;
            }
            else
            {
                return Convert.ToDouble(obj);
            }
        }

        public static decimal GetDecimalNullToZero(this object obj)
        {
            if (obj == null || obj == DBNull.Value || obj.ToString() == "")
            {
                return 0;
            }
            else
            {
                return Convert.ToDecimal(obj);
            }
        }

        public static bool GetBoolNullToFalse(this object obj)
        {
            if (obj is bool)
            {
                if (obj == null || obj == DBNull.Value || obj.ToString() == "")
                {
                    return false;
                }
                else
                {
                    return (bool)obj;
                }
            }
            else if (obj is string)
            {
                if (obj == null || obj == DBNull.Value || obj.ToString() == "")
                {
                    return false;
                }
                else
                {
                    return obj.ToString() == "Y" ? true : false;
                }
            }
            else
            {
                return false;
            }
        }

        public static T Min<T>(this T[] values)
        {
            return values.Min();
        }

        public static T Max<T>(this T[] values)
        {
            return values.Max();
        }

        public static string GetNullToEmpty(this object obj)
        {
            if (obj == null || obj == DBNull.Value)
            {
                return string.Empty;
            }
            else
            {
                return obj.ToString();
            }
        }

        public static string GetNullToNull(this object obj)
        {
            if (obj == null || obj == DBNull.Value)
            {
                return null;
            }
            else
            {
                return obj.ToString();
            }
        }

        public static DataTable GetEmptyDataTable(Dictionary<string, Type> columns)
        {
            DataTable dt = new DataTable();
            foreach (KeyValuePair<string, Type> columnName in columns)
            {
                dt.Columns.Add(columnName.Key, columnName.Value);
            }
            return dt;
        }

        public static DataTable GetAllDataTableInput(this DataTable dt, DataTable inputDataTable, Dictionary<string, string> columnMapping = null, Dictionary<string, string> expresstion = null)
        {
            foreach (DataRow row in inputDataTable.Rows)
            {
                DataRow newRow = dt.NewRow();

                foreach (DataColumn col in inputDataTable.Columns)
                {
                    if (dt.Columns.Contains(col.ColumnName))
                    {
                        newRow[col.ColumnName] = row[col];
                    }
                }

                if (columnMapping != null)
                {
                    foreach (KeyValuePair<string, string> kv in columnMapping)
                    {
                        if (dt.Columns.Contains(kv.Value) && inputDataTable.Columns.Contains(kv.Key))
                        {
                            newRow[kv.Value] = row[kv.Key];
                        }
                    }
                }

                dt.Rows.Add(newRow);
            }

            if (expresstion != null)
            {
                foreach (KeyValuePair<string, string> kv in expresstion)
                {
                    if (dt.Columns.Contains(kv.Key))
                    {
                        dt.Columns.Remove(kv.Key);
                    }

                    dt.Columns.Add(kv.Key, typeof(object), kv.Value);
                }
            }
            return dt;
        }

        public static long GetObjectMerorySize(object obj)
        {
            long size = 0;
            object o = new object();
            using (Stream s = new MemoryStream())
            {
                BinaryFormatter formatter = new BinaryFormatter();
                formatter.Serialize(s, obj);
                size = s.Length;
            }
            return size;
        }

        public static DataTable GetDictionaryToDataTable(this Dictionary<string, object> input)
        {
            DataTable dt = new DataTable("DATA");
            foreach (KeyValuePair<string, object> kv in input)
            {
                dt.Columns.Add(kv.Key, kv.Value.GetType());
            }
            DataRow newRow = dt.NewRow();
            foreach (KeyValuePair<string, object> kv in input)
            {
                newRow[kv.Key] = kv.Value;
            }
            dt.Rows.Add(newRow);
            return dt;
        }
        public static DataTable SetParaRename(this DataTable dt, string prefix = "V_")
        {
            foreach (DataColumn col in dt.Columns)
            {
                if (col.ColumnName != "R_CRUD")
                {
                    col.ColumnName = prefix + col.ColumnName;
                }
            }
            return dt;
        }

        public static bool isTablesTheSame(DataTable tbl1, DataTable tbl2)
        {
            if (tbl1.Rows.Count != tbl2.Rows.Count || tbl1.Columns.Count != tbl2.Columns.Count)
                return false;
            for (int i = 0; i < tbl1.Rows.Count; i++)
            {
                for (int c = 0; c < tbl1.Columns.Count; c++)
                {
                    if (!Equals(tbl1.Rows[i][c], tbl2.Rows[i][c]))
                        return false;
                }
            }
            return true;
        }

        /// <summary>
        /// DataTable Row1에 ColumnName, Row2에 컬럼명(한글) 일 경우 정리
        /// </summary>
        /// <param name="dt"></param>
        /// <returns></returns>
        public static DataTable GetExcelHeaderArrange(this DataTable dt)
        {
            if (dt.Rows.Count > 1)
            {
                foreach (DataColumn col in dt.Columns)
                {
                    col.ColumnName = dt.Rows[0][col].ToString();
                }
                dt.Rows[0].Delete();
                dt.AcceptChanges();
                dt.Rows[0].Delete();
                dt.AcceptChanges();
            }
            return dt;
        }

        public static DataTable ToDataTable(this IEnumerable ien)
        {
            DataTable dt = new DataTable();
            try
            {
                foreach (object obj in ien)
                {
                    Type t = obj.GetType();
                    PropertyInfo[] pis = t.GetProperties();
                    if (dt.Columns.Count == 0)
                    {
                        foreach (PropertyInfo pi in pis)
                        {
                            dt.Columns.Add(pi.Name, pi.PropertyType);
                        }
                    }
                    DataRow dr = dt.NewRow();
                    foreach (PropertyInfo pi in pis)
                    {
                        object value = pi.GetValue(obj, null);
                        dr[pi.Name] = value;
                    }
                    dt.Rows.Add(dr);
                }
            }
            catch (Exception ex)
            {
                throw ex;
            }
            return dt;
        }

        public static DataTable CopyToDataTable(this IEnumerable<DataRow> data)
        {
            DataTable dt = new DataTable();
            dt = data.First().Table.Clone();
            try
            {
                foreach (var row in data)
                {
                    dt.Rows.Add(row.ItemArray);
                }
            }
            catch (Exception ex)
            {
                throw ex;
            }
            return dt;
        }

        public static string ToDate10FromDate4(this string date)
        {
            try
            {
                if (date.Length != 4)
                {
                    throw new Exception("4자리 날짜가 아닙니다.");
                }
                else
                {
                    string ret = string.Empty;

                    string year = "201" + date.Left(1);
                    string month = string.Empty;
                    string day = date.Right(2);

                    switch (date.Substring(1, 1))
                    {
                        case "A":
                            month = "10";
                            break;
                        case "B":
                            month = "11";
                            break;
                        case "C":
                            month = "12";
                            break;
                        default:
                            month = date.Substring(1, 1);
                            break;
                    }

                    month = month.PadLeft(2, '0');

                    bool isDate = false;
                    DateTime returnDate;
                    isDate = DateTime.TryParse($"{year}-{month}-{day}", out returnDate);
                    if (isDate == false)
                    {
                        throw new Exception($"날짜 형식이 아닙니다.");
                    }
                    else
                    {
                        return returnDate.ToString("yyyy-MM-dd");
                    }
                }
            }
            catch (Exception ex)
            {
                throw new Exception(ex.Message);
            }
        }

    }

    public static class StringHandler
    {
        /// <summary>
        /// 문자열을 좌측으로 부터 특정 길이만큼 자른다.
        /// </summary>
        /// <param name="text">기준 문자열</param>
        /// <param name="textLen">자를 문자열의 길이</param>
        /// <returns>반환 문자열</returns>
        public static string Left(this string text, int textLen)
        {
            string convText;

            if (text.Length < textLen)
                textLen = text.Length;

            convText = text.Substring(0, textLen);
            return convText;
        }

        /// <summary>
        /// 문자열을 우측으로 부터 특정 길이만큼 자른다.
        /// </summary>
        /// <param name="text">기준 문자열</param>
        /// <param name="textLen">자를 문자열의 길이</param>
        /// <returns>반환 문자열</returns>
        public static string Right(this string text, int textLen)
        {
            string convText;
            if (text.Length < textLen)
                textLen = text.Length;

            convText = text.Substring(text.Length - textLen, textLen);
            return convText;
        }

        /// <summary>
        /// 문자열을 지정한 자리만큼 자른다.
        /// </summary>
        /// <param name="text">기준 문자열</param>
        /// <param name="startInt">문자열의 시작자리수</param>
        /// <param name="endInt">문자열의 종료자리수</param>
        /// <returns>반환 문자열</returns>
        public static string Mid(this string text, int startInt, int endInt)
        {
            string convText;
            if (startInt < text.Length || endInt < text.Length)
            {
                convText = text.Substring(startInt, endInt);
                return convText;
            }
            else
                return text;
        }

        /// <summary>
        /// 대소문자로 구성된 문자열에서 (_)를 대문자 앞에 추가하고 대문자로 변환하는 함수
        /// 첫번째 문자열은 제외된다.
        /// </summary>
        /// <param name="str"></param>
        /// <returns></returns>
        public static string UpperToUnderBarBySb(string str)
        {
            int i = 0;
            StringBuilder sb = new StringBuilder();
            foreach (Char c in str)
            {
                if (Char.IsUpper(c) && i > 0)
                    sb.Append("_");
                sb.Append(c);
                i++;
            }
            return sb.ToString().ToUpper();
        }

        public static string UpperToUnderBarByLinq(string str)
        {
            var converted = str.Select(x => Char.IsUpper(x) ? String.Concat("_", x) : x.ToString());
            var singleString = converted.Aggregate((a, b) => a + b);
            if (singleString.Left(1).Equals("_"))
                singleString = singleString.Right(singleString.Length - 1);
            return singleString.ToUpper();
        }

        public static string UpperToUnderBarByPattern(string str)
        {
            return Regex.Replace(str, "([a-z|0-9])([A-Z])", "$1_$2").ToUpper();
        }

        public static string NumberCheck(string num)
        {
            string retValue = "";
            Regex regex = new System.Text.RegularExpressions.Regex(@"^[0-9]{1,10}$");
            if ((num.Length > 1) && (num.Substring(0, 1) == "-"))
                retValue = "-";

            for (int i = 0; i < num.Length; i++)
            {
                Boolean ismatch = regex.IsMatch(num[i].ToString());
                if (!ismatch)
                    continue;
                else
                    retValue += num[i];
            }
            return retValue;
        }

        public enum StringAlignType
        {
            TEL
            , POST
            , BIZ
            , PERSON
            , DATE
            , DATEHAN
        }
        public static string StringAlign(StringAlignType type, string da)
        {
            string data = "";
            da = NumberCheck(da);
            if (da.Length < 5)
                return da;

            switch (type)
            {
                case StringAlignType.TEL:
                    if (da.Length > 8)
                    {
                        if (da.Substring(0, 2) == "02")
                        {
                            if (da.Length < 10)
                                data = da.Substring(0, 2) + "-" + da.Substring(2, 3) + "-" + da.Substring(5);
                            else
                                data = da.Substring(0, 2) + "-" + da.Substring(2, 4) + "-" + da.Substring(6);
                        }
                        else
                        {
                            if (da.Length < 11)
                                data = da.Substring(0, 3) + "-" + da.Substring(3, 3) + "-" + da.Substring(6);
                            else
                                data = da.Substring(0, 3) + "-" + da.Substring(3, 4) + "-" + da.Substring(7);
                        }
                    }
                    else
                        data = da.Substring(0, da.Length - 4) + "-" + da.Substring(da.Length - 4);
                    break;
                case StringAlignType.POST:
                    if (da.Length.Equals(6))
                        data = da.Substring(0, 3) + "-" + da.Substring(3);
                    else
                        data = da;
                    break;
                case StringAlignType.BIZ:
                    if (da.Length.Equals(10))
                        data = da.Substring(0, 3) + "-" + da.Substring(3, 2) + "-" + da.Substring(5);
                    else
                        data = da;
                    break;
                case StringAlignType.PERSON:
                    if (da.Length.Equals(13))
                        data = da.Substring(0, 6) + "-" + da.Substring(6);
                    else
                        data = da;
                    break;
                case StringAlignType.DATE:
                    if (da.Length.Equals(8))
                        data = da.Substring(0, 4) + "-" + da.Substring(4, 2) + "-" + da.Substring(6);
                    else if (da.Length.Equals(14))
                        data = da.Substring(0, 4) + "-" + da.Substring(4, 2) + "-" + da.Substring(6, 2) + " " + da.Substring(8, 2) + ":" + da.Substring(10, 2) + ":" + da.Substring(12);
                    else
                        data = da;
                    break;
                case StringAlignType.DATEHAN:
                    if (da.Length.Equals(8))
                        data = da.Substring(0, 4) + "년" + da.Substring(4, 2) + "월" + da.Substring(6) + "일";
                    else if (da.Length.Equals(14))
                        data = da.Substring(0, 4) + "년" + da.Substring(4, 2) + "월" + da.Substring(6, 2) + "일  " + da.Substring(8, 2) + "시" + da.Substring(10, 2) + "분" + da.Substring(12) + "초";
                    else
                        data = da;
                    break;

                default: break;
            }
            return data;
        }
    }

    public class ErrorHandler
    {
        //---------------------------------------------------------------------
        // Exception 개체를 전달받아서 InnerException의 메시지를 합한 다음
        // 그 내용으로 다시 오류를 발생시키는 메소드
        //---------------------------------------------------------------------
        public void RaiseException(Exception ex)
        {
            var exceptionMessage = new StringBuilder();
            do
            {
                exceptionMessage.Append(ex.Message);
                ex = ex.InnerException;
            }
            while (ex != null);
            throw new Exception(exceptionMessage.ToString());
        }

        //---------------------------------------------------------------------
        // Exception 개체를 전달받아서 InnerException의 메시지를 합한 다음
        // 그 내용을 반환하는 메소드
        //---------------------------------------------------------------------
        public string GetMessage(Exception ex)
        {
            var exceptionMessage = new StringBuilder();
            do
            {
                exceptionMessage.Append(ex.Message);
                ex = ex.InnerException;
            }
            while (ex != null);
            return exceptionMessage.ToString();
        }
    }

    public class BarcodeInfo
    {
        static char STX = (char)0x02;
        static char ETX = (char)0x03;

        /// <summary>
        /// 제품코드
        /// </summary>
        public string CustItemCd = string.Empty;
        /// <summary>
        /// 로트번호
        /// </summary>
        public string LotNo = string.Empty;
        /// <summary>
        /// 릴번호
        /// </summary>
        public int ReelNo = -1;
        /// <summary>
        /// 릴번호
        /// </summary>
        public string ReelNoString = string.Empty;
        /// <summary>
        /// 수량
        /// </summary>
        public int Qty = -1;
        /// <summary>
        /// 수량
        /// </summary>
        public string QtyString = string.Empty;
        /// <summary>
        /// 스캔된 바코드
        /// </summary>
        public string BarCode = string.Empty;
        /// <summary>
        /// 몰렉스설비코드
        /// </summary>
        public string MEquipCd = string.Empty;
        /// <summary>
        /// 날짜 4자리
        /// </summary>
        public string date = string.Empty;
        /// <summary>
        /// 날짜 10자리
        /// </summary>
        public string date10 = string.Empty;

        public BarcodeInfo(string barcode)
        {
            barcode = barcode.Replace(STX.ToString(), "").Replace(ETX.ToString(), "").Replace(Environment.NewLine, "").Replace("\r", "");
            BarCode = barcode;
            if (barcode.Length != 99)
            {
                throw new Exception("바코드가 99자리가 아닙니다.");
            }
            else
            {
                CustItemCd = barcode.Substring(6, 10);
                LotNo = barcode.Substring(32, 10);
                ReelNoString = barcode.Substring(45, 4);
                date = LotNo.Left(4);
                date10 = date.ToDate10FromDate4();

                bool isReel = false;
                isReel = int.TryParse(ReelNoString, out ReelNo);
                if (isReel == false)
                {
                    throw new Exception($"릴 번호 {ReelNoString} 숫자형이 아닙니다.");
                }

                QtyString = barcode.Substring(17, 12);
                bool isQty = false;
                isQty = int.TryParse(QtyString, out Qty);
                if (isReel == false)
                {
                    throw new Exception($"수량 {QtyString} 숫자형이 아닙니다.");
                }
                MEquipCd = LotNo.Substring(4, 2);
            }
        }
    }
}
